import re
from collections import defaultdict
from collections.abc import Callable, Iterable
from itertools import chain, groupby

import networkx as nx
from networkx import DiGraph
from rdflib import OWL, RDF, RDFS, SKOS, BNode, Graph, Literal, Namespace, URIRef
from rdflib.collection import Collection
from rdflib.namespace import split_uri
from thefuzz import fuzz, process

from cemento.rdf.constants import PREDICATES
from cemento.rdf.preprocessing import (
    clean_literal_string,
    get_abbrev_term,
    remove_term_names,
)


def construct_term_uri(
    prefix: str,
    abbrev_term: str,
    prefixes: dict[str, URIRef | Namespace],
) -> URIRef:
    ns_uri = prefixes[prefix]
    return URIRef(f"{ns_uri}{abbrev_term}")


def construct_literal(term: str, lang="en", datatype=None) -> Literal:
    return Literal(clean_literal_string(term), lang=lang, datatype=None)


def get_literal_lang_annotation(literal_term: str, default=None) -> str:
    return res[0] if (res := re.findall(r"@(\w+)", literal_term)) else default


def get_literal_data_type(
    literal_term: str,
    search_terms: dict[str, URIRef],
    score_cutoff=90,
) -> URIRef | None:
    search_key = res[0] if (res := re.findall(r"\^\^(\w+:\w+)", literal_term)) else None
    if search_key:
        datatype = substitute_term(
            [search_key], search_terms, score_cutoff=score_cutoff
        )
        return datatype
    return None


def substitute_term(
    search_keys: Iterable[str], search_terms: dict[str, URIRef], score_cutoff: int = 80
) -> URIRef:
    best_match, score = max(
        (
            result
            for search_key in search_keys
            if (
                result := process.extractOne(
                    search_key,
                    search_terms.keys(),
                    scorer=fuzz.token_sort_ratio,
                    score_cutoff=score_cutoff,
                )
            )
            is not None
        ),
        key=lambda x: x[1] if x is not None else -1,
        default=(None, -1),
    )
    return search_terms[best_match] if best_match else None


def get_class_terms(graph: DiGraph) -> set[URIRef]:
    class_terms = set()
    for subj, obj, data in graph.edges(data=True):
        predicate = data["label"]

        if predicate == RDFS.subClassOf:
            # we assume terms between RDFS.subClassOf is a class
            class_terms.update([subj, obj])

        if predicate == RDF.type:
            # we assume the object of an RDF.type is also a class
            class_terms.add(obj)

    return class_terms


def get_doms_ranges(graph: DiGraph) -> dict[URIRef, dict[str, list[URIRef]]]:
    doms_ranges = defaultdict(lambda: defaultdict(list))
    for domain_term, range_term, data in graph.edges(data=True):
        predicate_term = data.get("label", None)
        if predicate_term:
            doms_ranges[predicate_term]["domain"].append(domain_term)
            doms_ranges[predicate_term]["range"].append(range_term)
    return doms_ranges


def get_term_search_keys(term: str, inv_prefix: dict[URIRef, str]) -> list[str]:
    prefix, abbrev_term = get_abbrev_term(term)
    undo_camel_case_term = " ".join(
        re.findall(r"[A-Z]+(?=[A-Z][a-z]|\b)|[A-Z][a-z]+|[0-9]+", abbrev_term)
    )
    search_keys = [
        remove_term_names(term),
        f"{prefix}:{abbrev_term}",
        f"{prefix}:{undo_camel_case_term}",
    ]
    return [key.strip() for key in search_keys]


def get_term_search_result(
    term: URIRef,
    inv_prefixes: dict[URIRef | Namespace, str],
    search_terms: dict[str, URIRef],
) -> URIRef:
    ns, abbrev_term = split_uri(term)
    prefix = inv_prefixes[str(ns)]
    search_term = f"{prefix}:{abbrev_term}"
    if search_term in search_terms:
        return search_terms[search_term]
    return None


def iter_graph_terms(
    graph_terms: set[URIRef],
    term_function: Callable[[URIRef], None],
    filter_functions: Iterable[Callable[..., bool]] = None,
    invert_filter=False,
) -> list[any]:
    results = []
    for term in graph_terms:
        eval_value = True
        if filter_functions:
            eval_value = all(
                [filter_function(term) for filter_function in filter_functions]
            )
            if invert_filter:
                eval_value = not eval_value

        if eval_value:
            results.append(term_function(term))
    return results


def get_term_value(subj: URIRef, pred: URIRef, ref_rdf_graph: Graph):
    return ref_rdf_graph.value(subj, pred)


def bind_prefixes(rdf_graph: Graph, prefixes: dict[str, URIRef | Namespace]) -> Graph:
    for prefix, ns in prefixes.items():
        rdf_graph.bind(prefix, ns)
    return rdf_graph


def add_exact_matches(
    term: URIRef, match_properties: dict[URIRef, URIRef | None], rdf_graph: Graph
) -> Graph:
    # if the term is already imported from somewhere else
    # get the type and label if available and add to the ttl file
    for match_property, value in match_properties.items():
        if value:
            rdf_graph.add((term, match_property, value))

    # add an exact match to the ttl file for easier cross-referencing
    rdf_graph.add((term, SKOS.exactMatch, term))

    return rdf_graph


def add_labels(term: URIRef, labels: list[str], rdf_graph: Graph) -> Graph:
    # assume the first element of the labels is the actual label, others are alt-names
    if labels:
        rdf_graph.add((term, RDFS.label, Literal(labels[0])))
        if len(labels) > 1:
            for label in labels[1:]:
                rdf_graph.add((term, SKOS.altLabel, Literal(label)))
    return rdf_graph


def add_domains_ranges(
    term: URIRef,
    domains_ranges: dict[str, list[URIRef]],
    rdf_graph: Graph,
) -> Graph:
    predicate_domain = domains_ranges["domain"]
    predicate_range = domains_ranges["range"]

    for dom_or_range_term, term_dom_range in {
        RDFS.domain: predicate_domain,
        RDFS.range: predicate_range,
    }.items():
        if term_dom_range:
            # if there are more than one term, save the domain or range as a collection
            if len(term_dom_range) > 1:
                collection_node = BNode()
                Collection(rdf_graph, collection_node, term_dom_range)
                # create class that points to the collection
                collection_class = BNode()
                rdf_graph.add((collection_class, RDF.type, OWL.Class))
                # connect them all together
                # TODO: assume union for now but fix later
                rdf_graph.add((collection_class, OWL.unionOf, collection_node))
                rdf_graph.add((term, dom_or_range_term, collection_class))
            else:
                # if there is only one term, use that term directly
                rdf_graph.add((term, dom_or_range_term, term_dom_range[0]))


def get_aliases(rdf_graph: Graph) -> dict[URIRef, Literal]:
    label_tuples = list(
        chain(
            rdf_graph.subject_objects(RDFS.label),
            rdf_graph.subject_objects(SKOS.altLabel),
        )
    )
    sorted(label_tuples, key=lambda x: x[0])
    return {
        subj: [obj for _, obj in objs]
        for subj, objs in groupby(label_tuples, key=lambda x: x[0])
    }


def get_term_types(rdf_graph: Graph) -> dict[URIRef, URIRef]:
    return {subj: obj for subj, pred, obj in rdf_graph if pred == RDF.type}


def get_instances(
    rdf_graph: Graph, default_terms: set[URIRef], term_types: dict[URIRef, URIRef]
) -> set[URIRef]:
    return {
        subj
        for subj, pred, obj in rdf_graph
        if pred == RDF.type
        and obj not in default_terms
        and term_types[obj] == OWL.Class
    }


def get_classes(
    rdf_graph: Graph, default_terms: set[URIRef], term_types: dict[URIRef, URIRef]
) -> set[URIRef]:
    instance_superclasses = {
        subj
        for subj, pred, obj in rdf_graph
        if pred == RDF.type
        and obj not in default_terms
        and term_types[subj] == OWL.Class
    }
    subclass_terms = {
        term
        for subj, pred, obj in rdf_graph
        if pred == RDFS.subClassOf
        for term in (subj, obj)
    }
    return instance_superclasses | subclass_terms


def get_predicates(rdf_graph: Graph, default_terms: set[URIRef]) -> set[URIRef]:
    return {term for prop in PREDICATES for term in rdf_graph.subjects(RDF.type, prop)}


def get_graph_relabel_mapping(
    terms: URIRef,
    all_classes: set[URIRef],
    all_instances: set[URIRef],
    aliases: dict[URIRef, Literal],
    inv_prefix: dict[URIRef | Namespace, str],
) -> dict[URIRef, str]:
    rename_mapping = dict()
    for term in terms:
        ns, abbrev_term = split_uri(term)
        prefix = inv_prefix[ns]
        new_name = f"{prefix}:{abbrev_term}"
        if term in aliases and aliases[term]:
            if term in all_classes or term in all_instances:
                new_name += f" ({','.join(aliases[term])})"
            else:
                new_name = f"{prefix}:{aliases[term][0]}"
        rename_mapping[term] = new_name
    return rename_mapping


def get_graph(
    rdf_graph: Graph, all_predicates: set[URIRef], default_terms: set[URIRef]
) -> DiGraph:
    graph = DiGraph()
    for subj, pred, obj in rdf_graph:
        if (
            pred in all_predicates
            and subj not in default_terms
            and obj not in default_terms
        ):
            is_rank = pred in {RDF.type, RDFS.subClassOf}
            if is_rank:
                graph.add_edge(subj, obj, label=pred, is_rank=is_rank)
            else:
                graph.add_edge(obj, subj, label=pred, is_rank=is_rank)
    return graph


def rename_edges(graph: DiGraph, rename_mapping: dict[URIRef, str]) -> DiGraph:
    edge_rename_mapping = dict()
    graph = graph.copy()
    for subj, obj, data in graph.edges(data=True):
        pred = data["label"]
        new_edge_label = rename_mapping[pred]
        data.update({"label": new_edge_label})
        edge_rename_mapping[(subj, obj)] = data
    nx.set_edge_attributes(graph, edge_rename_mapping)
    return graph



import inspect
import json
import os
from collections.abc import Callable
from contextlib import contextmanager
from pathlib import Path

from networkx import DiGraph
from rdflib import RDFS, SKOS, Graph, Namespace, URIRef
from rdflib.namespace import split_uri
from collections.abc import Iterable

def iter_diagram_terms(
    graph: DiGraph,
    term_function: (
        Callable[[str | URIRef, bool], str | URIRef]
        | Callable[[str | URIRef], str | URIRef]
    ),
) -> list[any]:
    results = []
    for subj, obj, data in graph.edges(data=True):
        pred = data['label']
        for term in (subj, pred, obj):
            is_predicate = term == pred
            if "is_predicate" in inspect.signature(term_function).parameters:
                result = term_function(term, is_predicate)
            else:
                result = term_function(term)
            results.append(result)
    return results

def get_diagram_terms_iter_with_pred(graph: DiGraph) -> Iterable[str]:
    return ((term, term == data['label']) for subj, obj, data in graph.edges(data=True) for term in (subj, data['label'], obj))

def get_diagram_terms_iter(graph: DiGraph) -> Iterable[str]:
    return (term for subj, obj, data in graph.edges(data=True) for term in (subj, data['label'], obj))

def get_ttl_file_iter(folder_path: str | Path) -> Iterable[Path]:
    return (get_ttl_graph(file_path) for file in os.scandir(folder_path) if (file_path := Path(file.path)).suffix == '.ttl')

def iterate_ttl_graphs(
    folder_path: str, graph_function: Callable[[Graph], any]
) -> list[any]:
    results = []
    for file in os.scandir(folder_path):
        file_path = Path(file.path)
        if file_path.suffix == ".ttl":
            with read_ttl(file_path) as graph:
                result = graph_function(graph)
                results.append(result)
    return results

def get_ttl_graph(file_path: str | Path) -> Graph | None:
    with read_ttl(file_path) as graph:
        return graph

@contextmanager
def read_ttl(file_path: str | Path) -> Graph:
    rdf_graph = Graph()
    try:
        rdf_graph.parse(file_path, format="turtle")
        yield rdf_graph
    finally:
        rdf_graph.close()


def read_prefixes_from_json(file_path: str) -> dict[str, URIRef]:
    with open(file_path, "r") as f:
        prefixes = json.load(f)
        return prefixes


def get_search_terms_from_defaults(
    default_namespace_prefixes: dict[str, Namespace],
) -> dict[str, URIRef]:
    search_terms = dict()
    for prefix, ns in default_namespace_prefixes.items():
        for term in dir(ns):
            if isinstance(term, URIRef):
                _, name = split_uri(term)
                search_terms[f"{prefix}:{name}"] = term
    return search_terms


def read_prefixes_from_graph(rdf_graph: Graph) -> dict[str, str]:
    return {prefix: str(ns) for prefix, ns in rdf_graph.namespaces()}


def get_search_terms_from_graph(
    rdf_graph: Graph, inv_prefixes: dict[str, str]
) -> dict[str, URIRef]:
    search_terms = dict()
    all_terms = set()
    for subj, pred, obj in rdf_graph:
        all_terms.update([subj, pred, obj])

        if pred == RDFS.label or pred == SKOS.altLabel:
            ns, _ = split_uri(subj)
            prefix = inv_prefixes[ns]
            search_terms[f"{prefix}:{str(obj)}"] = subj

    for term in all_terms:
        if isinstance(term, URIRef):
            is_literal = False
            try:
                ns, abbrev_term = split_uri(term)
            except ValueError:
                is_literal = not is_literal

            if not is_literal:
                prefix = inv_prefixes[str(ns)]
                search_terms[f"{prefix}:{abbrev_term}"] = term

    return search_terms



from pathlib import Path

import networkx as nx
from networkx import DiGraph
from rdflib import DCTERMS, OWL, RDF, RDFS, SKOS, URIRef

from cemento.rdf.io import read_ttl
from cemento.rdf.transforms import (
    get_aliases,
    get_classes,
    get_graph,
    get_graph_relabel_mapping,
    get_instances,
    get_predicates,
    get_term_types,
    rename_edges,
)


def convert_ttl_to_graph(input_path: str | Path) -> DiGraph:
    default_namespaces = [RDF, RDFS, OWL, DCTERMS, SKOS]
    default_namespace_prefixes = ["rdf", "rdfs", "owl", "dcterms", "skos"]

    with read_ttl(input_path) as rdf_graph:

        prefixes = {prefix: ns for prefix, ns in rdf_graph.namespaces()}
        prefixes.update(
            {
                prefix: ns
                for prefix, ns in zip(default_namespace_prefixes, default_namespaces)
            }
        )

        inv_prefix = {str(value): key for key, value in prefixes.items()}

        default_terms = {
            term
            for ns in default_namespaces
            for term in dir(ns)
            if isinstance(term, URIRef)
        }

        term_types = get_term_types(rdf_graph)
        all_classes = get_classes(rdf_graph, default_terms, term_types)
        all_instances = get_instances(rdf_graph, default_terms, term_types)
        all_predicates = get_predicates(rdf_graph, default_terms)
        all_predicates.update([RDF.type, RDFS.subClassOf])
        graph = get_graph(rdf_graph, all_predicates, default_terms)

        all_terms = all_classes | all_instances | all_predicates | default_terms
        aliases = get_aliases(rdf_graph)
        rename_terms = get_graph_relabel_mapping(
            all_terms, all_classes, all_instances, aliases, inv_prefix
        )
        graph = nx.relabel_nodes(graph, rename_terms)
        graph = rename_edges(graph, rename_terms)
        return graph



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from math import atan2, pi


class DiagramKey(Enum):
    TERM_ID = "term_id"
    LABEL = "label"


SHAPE_WIDTH = 200
SHAPE_HEIGHT = 80
x_padding = 10
y_padding = 20


def get_timestamp_str():
    return f"{datetime.now():%Y-%m-%dT%H:%M:%S.%fZ}"


class DiagramObject:
    pass


@dataclass
class DiagramInfo(DiagramObject):
    diagram_name: int
    diagram_id: int
    modify_date: str = field(default_factory=get_timestamp_str)
    grid_dx: int = 1600
    grid_dy: int = 850
    grid_size: int = 10
    page_width: int = 1100
    page_height: int = 850
    diagram_content: int = None
    template_key: str = "scaffold"


class ConnectorType(Enum):
    RANK_CONNECTOR = "rank"
    PROPERTY_CONNECTOR = "property"


@dataclass
class Connector(DiagramObject):
    connector_id: str
    source_id: str
    target_id: str
    connector_label_id: str
    connector_val: str
    rel_x_pos: float = 0
    rel_y_pos: float = 0
    start_pos_x: float = 0.5
    start_pos_y: float = 1
    end_pos_x: float = 0.5
    end_pos_y: float = 0
    is_dashed: bool = 0
    is_curved: bool = 0
    template_key: str = "connector"

    @staticmethod
    def compute_dynamic_position(
        source_shape_x: float,
        source_shape_y: float,
        target_shape_x: float,
        target_shape_y: float,
    ) -> tuple[float, float, float, float]:
        crit_angle = abs(atan2(SHAPE_HEIGHT, SHAPE_WIDTH))
        # apple around banana
        current_angle = atan2(target_shape_y - source_shape_y, target_shape_x - source_shape_x)
        match current_angle:
            # left
            case angle if (
                pi - crit_angle
            ) <= angle <= 180 or -180 <= angle < crit_angle:
                return (0, 0.5, 1, 0.5)
            # top
            case angle if -(pi - crit_angle) <= angle < -crit_angle or angle == 2 * pi:
                return (0.5, 0, 0.5, 1)
            # right
            case angle if -crit_angle <= angle < crit_angle:
                return (0, 0.5, 1, 0.5)
            # bottom
            case angle if crit_angle <= angle < (pi - crit_angle):
                return (0.5, 1, 0.5, 0)
            case _:
                return (0, 0, 0, 0)

    def resolve_position(
        self,
        connector_type: str | ConnectorType = None,
        source_shape_pos: tuple[float, float] = None,
        target_shape_pos: tuple[float, float] = None,
    ) -> None:
        match connector_type:
            case ConnectorType.RANK_CONNECTOR | ConnectorType.RANK_CONNECTOR.value:
                self.start_pos_x = 0.5
                self.start_pos_x = 1
                self.end_pos_x = 0.5
                self.end_pos_y = 0
                return
            case (
                ConnectorType.PROPERTY_CONNECTOR
                | ConnectorType.PROPERTY_CONNECTOR.value
            ):
                self.start_pos_x, self.start_pos_y, self.end_pos_x, self.end_pos_y = (
                    Connector.compute_dynamic_position(
                        *source_shape_pos, *target_shape_pos
                    )
                )
            case _:
                return

@dataclass
class GhostConnector(Connector):
    is_curved: bool = 1
    is_dashed: bool = 1

@dataclass
class Shape:
    shape_id: str
    shape_content: str
    fill_color: str
    x_pos: float
    y_pos: float
    shape_width: int
    shape_height: int
    template_key: str = "shape"



from cemento.rdf.drawio_to_turtle import convert_drawio_to_ttl

INPUT_PATH = ""
OUTPUT_PATH = ""
ONTO_PATH = "" # this path points to the folder containing all the ttl files you want to reference (optional)
PREFIXES_PATH = "" # this path points to a prefixes file with custom prefix assignments (optional)

if __name__ == "__main__":
    convert_drawio_to_ttl(INPUT_PATH, OUTPUT_PATH, ONTO_PATH, PREFIXES_PATH)


from dataclasses import dataclass, field
from datetime import datetime

SHAPE_WIDTH = 200
SHAPE_HEIGHT  = 80
x_padding = 10
y_padding = 20

def get_timestamp_str():
    return f"{datetime.now():%Y-%m-%dT%H:%M:%S.%fZ}"

@dataclass
class DiagramInfo:
    modify_date: str = field(default_factory=get_timestamp_str)
    diagram_name: int
    diagram_id: int
    grid_dx: int = 1600
    grid_dy: int = 850
    grid_size: int = 10
    page_width: int = 1100
    page_height: int = 850
    diagram_content: int = None

@dataclass
class Connector:
    id: str
    source_id: str
    target_id: str
    connector_label_id: str
    connector_content: str
    is_rank: str
    start_x: str
    start_y: str
    end_x: str
    end_y: str

@dataclass
class Shape:
    shape_id: str
    shape_content: str
    fill_color: str
    x_pos: int
    y_pos: int
    shape_width: int
    shape_height: int



from pathlib import Path

import networkx as nx
from defusedxml import ElementTree as ET
from networkx import DiGraph
from thefuzz import process

from cemento.draw_io.constants import DiagramKey
from cemento.draw_io.preprocessing import clean_term


def parse_elements(file_path: str | Path) -> dict[str, dict[str, any]]:
    # parse elements
    tree = ET.parse(file_path)
    root = tree.getroot()
    root = next(tree.iter("root"))

    elements = dict()
    all_cells = [child for child in root.findall("mxCell")]  # only add base level cells

    for cell in all_cells:
        cell_attrs = dict()
        cell_attrs.update(cell.attrib)

        # add nested position attributes if any
        nested_attrs = dict()
        for subcell in cell.findall("mxGeometry"):
            nested_attrs.update(subcell.attrib)
        cell_attrs.update(nested_attrs)

        # add style attributes as data attributes
        if "style" in cell.attrib:
            style_terms = dict()
            style_tags = []
            for style_attrib_string in cell.attrib["style"].split(";"):
                style_term = style_attrib_string.split("=")
                if len(style_term) > 1:  # for styles with values
                    key, value = style_term
                    style_terms[key] = value
                elif style_term[0]:  # for style tags or names
                    style_tags.append(style_term[0])
            cell_attrs.update(style_terms)
            cell_attrs["tags"] = style_tags
            del cell_attrs["style"]  # remove style to prevent redundancy

        cell_id = cell.attrib["id"]
        del cell_attrs["id"]  # remove id since it is now used as a key
        elements[cell_id] = (
            cell_attrs  # set dictionary of cell key and attribute dictionary
        )

    return elements


def extract_elements(
    elements: dict[str, dict[str, any]],
) -> tuple[set[dict[str, any], set[str]]]:
    # read edges
    term_ids, rel_ids = set(), set()
    for element, data in elements.items():
        # if the vertex attribute is 1 and edgeLabel is not in tags, it is a term (shape)
        if (
            "vertex" in data
            and data["vertex"] == "1"
            and (
                "tags" not in data
                or ("tags" in data and "edgeLabel" not in data["tags"])
            )
        ):
            term_ids.add(element)
        # if an element has an edgeLabel tag, it is a relationship (connection)
        elif "tags" in data and "edgeLabel" in data["tags"]:
            rel_val = data["value"]
            rel_id = data["parent"]
            elements[rel_id]["value"] = rel_val
            rel_ids.add(rel_id)
        # if an element has value, source and target, it is also a relationship (connection)
        elif "value" in data and "source" in data and "target" in data:
            rel_val = data["value"]
            rel_id = element
            rel_ids.add(rel_id)
    return term_ids, rel_ids


def generate_graph(
    elements: dict[str, dict[str, any]],
    term_ids: set[str],
    relationship_ids: set[str],
    rank_terms: set[str],
    inverted_rank_arrow: bool = False,
) -> DiGraph:
    # for identified connectors, extract relationship information
    graph = nx.DiGraph()
    # add all terms
    for term_id in term_ids:
        term = elements[term_id]["value"]
        graph.add_node(term_id, term_id=term_id, label=clean_term(term))

    # add all relationships
    for rel_id in relationship_ids:
        subj_id = elements[rel_id]["source"]
        obj_id = elements[rel_id]["target"]
        pred = clean_term(elements[rel_id]["value"])

        matched_rank_pred, score = process.extractOne(
            pred.lower(), [rank_term.lower() for rank_term in rank_terms]
        )
        is_rank = score > 85
        pred = matched_rank_pred if is_rank else pred

        # arrow conventions are inverted for rank relationships, flip assignments to conform
        if inverted_rank_arrow and is_rank:
            temp = subj_id
            subj_id = obj_id
            obj_id = temp

        graph.add_edge(
            subj_id,
            obj_id,
            label=pred,
            pred_id=rel_id,
            is_rank=is_rank,
            is_predicate=True,
        )

    return graph


def relabel_graph_nodes(
    graph: DiGraph, new_attr_label=DiagramKey.TERM_ID.value
) -> DiGraph:
    node_info = nx.get_node_attributes(graph, new_attr_label)
    relabel_mapping = {
        current_node_label: node_info[current_node_label]
        for current_node_label in graph.nodes
    }
    return nx.relabel_nodes(graph, relabel_mapping)



from pathlib import Path

from defusedxml import ElementTree as ET


def get_diagram_headers(file_path: str | Path) -> dict[str, str]:
    # retrieve diagram headers
    tree = ET.parse(file_path)
    root = tree.getroot()
    graph_model_tag = next(root.iter("mxGraphModel"))
    diagram_tag = next(root.iter("diagram"))
    diagram_headers = {
        "modify_date": (
            root.attrib["modified"] if "modified" in root.attrib else "July 1, 2024"
        ),
        "diagram_name": diagram_tag.attrib["name"],
        "diagram_id": diagram_tag.attrib["id"],
        "grid_dx": graph_model_tag.attrib["dx"] if "dx" in graph_model_tag else 0,
        "grid_dy": graph_model_tag.attrib["dy"] if "dy" in graph_model_tag else 0,
        "grid_size": int(graph_model_tag.attrib["gridSize"]),
        "page_width": int(graph_model_tag.attrib["pageWidth"]),
        "page_height": int(graph_model_tag.attrib["pageHeight"]),
    }

    return diagram_headers



import importlib.resources as pkg_resources
import os
from dataclasses import asdict
from pathlib import Path
from string import Template
from uuid import uuid4

import networkx as nx
from networkx import DiGraph

from cemento.draw_io.constants import (
    Connector,
    DiagramInfo,
    DiagramObject,
    GhostConnector,
    Shape,
)

INPUT_PATH = "/Users/gabbython/dev/sdle/CEMENTO/sandbox/SyncXrayResult_graph.drawio"


def get_ranked_subgraph(graph: DiGraph) -> DiGraph:
    ranked_subgraph = graph.copy()
    ranked_subgraph.remove_edges_from(
        [
            (subj, obj)
            for subj, obj, data in graph.edges(data=True)
            if not data["is_rank"]
        ]
    )
    return ranked_subgraph


def get_subgraphs(graph: DiGraph) -> list[DiGraph]:
    subgraphs = nx.weakly_connected_components(graph)
    return [graph.subgraph(subgraph_nodes).copy() for subgraph_nodes in subgraphs]


def get_graph_root_nodes(graph: DiGraph) -> list[any]:
    return [node for node in graph.nodes if graph.in_degree(node) == 0]


def split_multiple_inheritances(
    graph: DiGraph,
) -> tuple[list[DiGraph], list[tuple[any, any]]]:
    # create a dummy graph and connect root nodes to a dummy node
    dummy_graph = graph.copy()
    root_nodes = get_graph_root_nodes(dummy_graph)
    dummy_graph.add_edges_from(("dummy", root_node) for root_node in root_nodes)
    fork_nodes = [
        node
        for node in nx.dfs_postorder_nodes(dummy_graph, source="dummy")
        if len(list(dummy_graph.predecessors(node))) > 1
    ]

    if len(fork_nodes) == 0:
        return [graph], []

    fork_levels = {
        node: nx.shortest_path_length(dummy_graph, source="dummy", target=node)
        for node in fork_nodes
    }
    sorted(fork_nodes, key=lambda x: fork_levels[x])
    dummy_graph.remove_node("dummy")

    diamond_heads = set()
    for root in root_nodes:
        for fork in fork_nodes:
            paths = list(nx.all_simple_paths(dummy_graph, source=root, target=fork))
            if len(paths) > 1:
                diamond_heads.add(paths[0][0])

    severed_links = list()
    for fork_node in fork_nodes:
        fork_predecessors = list(dummy_graph.predecessors(fork_node))
        edges_to_cut = [
            (predecessor, fork_node) for predecessor in fork_predecessors[1:]
        ]
        dummy_graph.remove_edges_from(edges_to_cut)
        severed_links.extend(edges_to_cut)

    for diamond_head in diamond_heads:
        diamond_successors = list(dummy_graph.successors(diamond_head))
        edges_to_cut = [
            (diamond_head, successor) for successor in diamond_successors[1:]
        ]
        dummy_graph.remove_edges_from(edges_to_cut)
        severed_links.extend(edges_to_cut)

    subtrees = get_subgraphs(dummy_graph)
    return subtrees, severed_links


def compute_grid_allocations(tree: DiGraph, root_node: any) -> DiGraph:
    tree = tree.copy()
    for node in tree.nodes:
        set_reserved_x = 1 if tree.out_degree(node) == 0 else 0
        tree.nodes[node]["reserved_x"] = set_reserved_x
        tree.nodes[node]["reserved_y"] = 1

    for node in reversed(list(nx.bfs_tree(tree, root_node))):
        if len(nx.descendants(tree, node)) > 0:
            max_reserved_y = 0
            for child in tree.successors(node):
                new_reserved_x = (
                    tree.nodes[node]["reserved_x"] + tree.nodes[child]["reserved_x"]
                )
                max_reserved_y = max(max_reserved_y, tree.nodes[node]["reserved_y"])
                tree.nodes[node]["reserved_x"] = new_reserved_x
            new_reserved_y = max_reserved_y + tree.nodes[node]["reserved_y"]
            tree.nodes[node]["reserved_y"] = new_reserved_y
    return tree


def get_tree_size(tree: DiGraph) -> tuple[int, int]:
    tree_size_x = max(nx.get_node_attributes(tree, "reserved_x").values())
    tree_size_y = max(nx.get_node_attributes(tree, "reserved_y").values())
    return (tree_size_x, tree_size_y)


def compute_draw_positions(
    tree: DiGraph, root_node: any, horizontal_tree: bool = False
) -> DiGraph:
    tree = tree.copy()
    nodes_drawn = set()
    for level, nodes_in_level in enumerate(nx.bfs_layers(tree, root_node)):
        for node in nodes_in_level:
            tree.nodes[node]["draw_y"] = level
            nodes_drawn.add(node)

    tree.nodes[root_node]["cursor_x"] = 0
    for node in nx.dfs_preorder_nodes(tree, root_node):
        offset_x = 0
        cursor_x = tree.nodes[node]["cursor_x"]

        for child in tree.successors(node):
            child_cursor_x = cursor_x + offset_x
            tree.nodes[child]["cursor_x"] = child_cursor_x
            offset_x += tree.nodes[child]["reserved_x"]

        tree.nodes[node]["draw_x"] = (2 * cursor_x + tree.nodes[node]["reserved_x"]) / 2
        nodes_drawn.add(node)

    remaining_nodes = tree.nodes - nodes_drawn
    for node in remaining_nodes:
        reserved_x = tree.nodes[node]["reserved_x"]
        reserved_y = tree.nodes[node]["reserved_y"]
        cursor_x = tree.nodes[node]["cursor_x"] if "cursor_x" in tree.nodes[node] else 0
        cursor_y = tree.nodes[node]["cursor_y"] if "cursor_y" in tree.nodes[node] else 0

        tree.nodes[node]["draw_x"] = cursor_x + reserved_x
        tree.nodes[node]["draw_y"] = cursor_y + reserved_y

    if horizontal_tree:
        for node in tree.nodes:
            draw_x = tree.nodes[node]["draw_x"]
            draw_y = tree.nodes[node]["draw_y"]

            tree.nodes[node]["draw_y"] = draw_x
            tree.nodes[node]["draw_x"] = draw_y

    return tree


def translate_coords(
    x_pos: float, y_pos: float, origin_x: float = 0, origin_y: float = 0
) -> tuple[int, int]:
    rect_width = 200
    rect_height = 80
    x_padding = 5
    y_padding = 20

    grid_x = rect_width * 2 + x_padding
    grid_y = rect_height * 2 + y_padding

    return ((x_pos + origin_x) * grid_x, (y_pos + origin_y) * grid_y)


def generate_shapes(
    graph: DiGraph,
    diagram_uid: str,
    offset_x: int = 0,
    offset_y: int = 0,
    idx_start: int = 0,
) -> list[Shape]:
    shapes = []
    for idx, node in enumerate(graph.nodes):
        node_content = node.replace('"', "&quot;")
        entity_id = f"{diagram_uid}-{idx + idx_start}"
        shape_pos_x = graph.nodes[node]["draw_x"] + offset_x
        shape_pos_y = graph.nodes[node]["draw_y"] + offset_y
        shape_pos_x, shape_pos_y = translate_coords(shape_pos_x, shape_pos_y)
        shapes.append(
            Shape(
                shape_id=entity_id,
                shape_content=node_content,
                fill_color="#f2f3f4",
                x_pos=shape_pos_x,
                y_pos=shape_pos_y,
                shape_width=200,
                shape_height=80,
            )
        )
    return shapes


def get_template_files() -> dict[str, str | Path]:
    current_file_folder = Path(__file__)
    # retrieve the template folder from the grandparent directory
    template_path = current_file_folder.parent.parent.parent / "templates"

    if not template_path.exists():
        template_path = pkg_resources.files("cemento").joinpath("templates")

    template_files = [
        Path(file) for file in os.scandir(template_path) if file.name.endswith(".xml")
    ]
    template_dict = dict()
    for file in template_files:
        with open(file, "r") as f:
            template = Template(f.read())
            template_dict[file.stem] = template
    return template_dict


def generate_diagram_content(
    diagram_name: str, diagram_uid: str, *diagram_objects: list[DiagramObject]
) -> str:
    diagram_info = DiagramInfo(diagram_name, diagram_uid)

    diagram_content = ""
    templates = get_template_files()
    diagram_content += "".join(
        [
            templates[obj.template_key].substitute(asdict(obj))
            for objects in diagram_objects
            for obj in objects
        ]
    )
    diagram_info.diagram_content = diagram_content
    write_content = templates[diagram_info.template_key].substitute(
        asdict(diagram_info)
    )
    return write_content


def get_shape_ids(shapes: list[Shape]) -> dict[str, Shape]:
    return {shape.shape_content: shape.shape_id for shape in shapes}


def get_shape_positions(shapes: list[Shape]) -> dict[str, tuple[float, float]]:
    return {shape.shape_content: (shape.x_pos, shape.y_pos) for shape in shapes}


def draw_tree(
    graph: DiGraph,
    diagram_output_path: str | Path,
    translate_x: int = 0,
    translate_y: int = 0,
    horizontal_tree: bool = False,
) -> None:
    diagram_output_path = Path(diagram_output_path)
    ranked_graph = get_ranked_subgraph(graph)
    ranked_graph = ranked_graph.reverse(copy=True)
    ranked_subtrees = get_subgraphs(ranked_graph)
    split_subtrees, severed_links = zip(
        *map(split_multiple_inheritances, ranked_subtrees)
    )
    ranked_subtrees = [tree for trees in split_subtrees for tree in trees]
    severed_links = [edge for edges in severed_links for edge in edges]

    shapes = []
    ranked_subtrees = map(
        lambda subtree: compute_grid_allocations(
            subtree, get_graph_root_nodes(subtree)[0]
        ),
        ranked_subtrees,
    )

    ranked_subtrees = map(
        lambda subtree: compute_draw_positions(
            subtree, get_graph_root_nodes(subtree)[0]
        ),
        ranked_subtrees,
    )

    diagram_uid = str(uuid4()).split("-")[-1]
    offset_x, offset_y = 0, 0
    entity_idx_start = 0
    ranked_subtrees = list(ranked_subtrees)
    for subtree in ranked_subtrees:

        new_shapes = generate_shapes(
            subtree,
            diagram_uid,
            offset_x=offset_x,
            offset_y=offset_y,
            idx_start=entity_idx_start,
        )

        tree_size_x, tree_size_y = get_tree_size(subtree)
        offset_x += tree_size_x
        offset_y += tree_size_y

        if horizontal_tree:
            offset_x = 0
        else:
            offset_y = 0

        shapes.extend(new_shapes)
        entity_idx_start += len(new_shapes)

    new_shape_ids = get_shape_ids(shapes)
    connectors = []
    connector_idx = entity_idx_start + 1
    for subtree in ranked_subtrees:
        new_connectors = []
        for subj, obj, data in subtree.edges(data=True):
            pred = data["label"].replace('"', "").strip()
            new_connectors.append(
                Connector(
                    connector_id=f"{diagram_uid}-{connector_idx}",
                    source_id=new_shape_ids[subj],
                    target_id=new_shape_ids[obj],
                    connector_label_id=f"{diagram_uid}-{connector_idx + 1}",
                    connector_val=pred,
                )
            )
            connector_idx += 2
        entity_idx_start += len(new_connectors)
        connectors.extend(new_connectors)

    predicate_connectors = []
    shape_positions = get_shape_positions(shapes)
    property_edges = (
        (subj, obj, data["label"])
        for subj, obj, data in graph.edges(data=True)
        if not data["is_rank"]
    )
    for subj, obj, pred in property_edges:
        new_connector = GhostConnector(
            connector_id=f"{diagram_uid}-{connector_idx}",
            source_id=new_shape_ids[subj],
            target_id=new_shape_ids[obj],
            connector_label_id=f"{diagram_uid}-{connector_idx + 1}",
            connector_val=pred,
        )
        predicate_connectors.append(new_connector)
        new_connector.resolve_position(
            "property", shape_positions[subj], shape_positions[obj]
        )
        connector_idx += 2
    write_content = generate_diagram_content(
        diagram_output_path.stem, diagram_uid, connectors, predicate_connectors, shapes
    )
    with open(diagram_output_path, "w") as write_file:
        write_file.write(write_content)



from cemento.draw_io.read_diagram import read_drawio
from pathlib import Path
from cemento.rdf.graph_to_turtle import convert_graph_to_ttl

def convert_drawio_to_ttl(
    file_path: str | Path,
    output_path: str | Path,
    onto_ref_folder: str | Path,
    prefixes_path: str | Path,
) -> None:
    convert_graph_to_ttl(
        read_drawio(file_path), output_path, onto_ref_folder, prefixes_path
    )


from pathlib import Path

from cemento.draw_io.write_diagram import draw_tree
from cemento.rdf.turtle_to_graph import convert_ttl_to_graph


def convert_ttl_to_drawio(
    input_path: str | Path, output_path: str | Path, horizontal_tree: bool = False
):
    graph = convert_ttl_to_graph(input_path)
    draw_tree(graph, output_path, horizontal_tree=horizontal_tree)



import re
from collections import defaultdict

import tldextract
from rdflib import Graph, Namespace, URIRef
from rdflib.namespace import split_uri


def merge_dictionaries(dict_list: list[dict[any, any]]) -> dict[any, any]:
    return {key: value for each_dict in dict_list for key, value in each_dict.items()}


def clean_literal_string(literal_term: str) -> str:
    new_literal_term = literal_term.strip().replace('"', "")
    new_literal_term = re.sub(r"@\w+", "", new_literal_term)
    new_literal_term = re.sub(r"\^\^\w+:\w+", "", new_literal_term)
    return new_literal_term


def generate_residual_prefixes(
    rdf_graph: Graph, inv_prefixes: dict[Namespace | URIRef, str]
):
    new_prefixes = defaultdict(list)
    new_prefix_namespaces = set()
    for subj, pred, obj in rdf_graph:
        for term in [subj, pred, obj]:
            if isinstance(term, URIRef):
                try:
                    ns, abbrev = split_uri(term)
                except ValueError:
                    ns = term
                if ns not in inv_prefixes:
                    new_prefix_namespaces.add(str(ns))
    gns_idx = 0
    for ns in new_prefix_namespaces:
        url_extraction = tldextract.extract(ns)
        new_prefix = res[-1] if (res := re.findall(r"\w+", ns)) else ""
        if url_extraction.suffix and new_prefix in url_extraction.suffix.split("."):
            new_prefix = url_extraction.domain
        new_prefix = re.sub(r"[^a-zA-Z0-9]", "", new_prefix)
        if not new_prefix or new_prefix.isdigit():
            new_prefix = f"gns{gns_idx}"
            gns_idx += 1
        new_prefixes[new_prefix].append(ns)

    return_prefixes = dict()
    for prefix, namespaces in new_prefixes.items():
        if len(namespaces) > 1:
            for idx, ns in enumerate(namespaces):
                return_prefixes[f"{prefix}{idx+1}"] = ns
        else:
            return_prefixes[prefix] = namespaces[0]

    return return_prefixes


def remove_term_names(term: str) -> str:
    match = re.search(r"^([^(]*)", term)
    return match.group(1).strip() if match else term


def get_term_aliases(term: str) -> list[str]:
    match = re.search(r"\(([^)]*)\)", term)
    if match:
        alt_term_string = match.group(1)
        alt_term_string = alt_term_string.split(",")
        return [term.strip() for term in alt_term_string]
    return []


def get_abbrev_term(
    term: str, is_predicate=False, default_prefix="mds"
) -> tuple[str, str]:
    prefix = default_prefix
    abbrev_term = term
    strict_camel_case = False

    term = remove_term_names(term)
    if ":" in term:
        prefix, abbrev_term = term.split(":")

    if is_predicate:
        abbrev_term = abbrev_term.replace("_", " ")
        strict_camel_case = not strict_camel_case

    # if the term is a class, use upper camel case / Pascal case
    abbrev_term = "".join(
        [
            f"{word[0].upper()}{word[1:] if len(word) > 1 else ''}"
            for word in abbrev_term.split()
        ]
    )

    if strict_camel_case and term[0].islower():
        abbrev_term = (
            f"{abbrev_term[0].lower()}{abbrev_term[1:] if len(abbrev_term) > 1 else ''}"
        )

    return prefix, abbrev_term



from functools import partial, reduce
from itertools import filterfalse
from pathlib import Path

import networkx as nx
import rdflib
from networkx import DiGraph
from rdflib import DCTERMS, OWL, RDF, RDFS, SKOS

from cemento.rdf.filters import term_in_search_results, term_not_in_default_namespace
from cemento.rdf.io import (
    get_diagram_terms_iter,
    get_diagram_terms_iter_with_pred,
    get_search_terms_from_defaults,
    get_search_terms_from_graph,
    get_ttl_file_iter,
    iter_diagram_terms,
    iterate_ttl_graphs,
    read_prefixes_from_graph,
    read_prefixes_from_json,
)
from cemento.rdf.preprocessing import (
    generate_residual_prefixes,
    get_abbrev_term,
    get_term_aliases,
    merge_dictionaries,
)
from cemento.rdf.transforms import (
    add_domains_ranges,
    add_exact_matches,
    add_labels,
    bind_prefixes,
    construct_literal,
    construct_term_uri,
    get_class_terms,
    get_doms_ranges,
    get_literal_data_type,
    get_term_search_keys,
    get_term_value,
    substitute_term,
    get_literal_lang_annotation
)


def convert_graph_to_ttl(
    graph: DiGraph,
    output_path: str | Path,
    onto_ref_folder: str | Path = None,
    prefixes_path: str | Path = None,
) -> None:
    default_namespaces = [RDF, RDFS, OWL, DCTERMS, SKOS]
    default_namespace_prefixes = ["rdf", "rdfs", "owl", "dcterms", "skos"]

    prefixes = dict()
    if prefixes_path:
        prefixes = read_prefixes_from_json(prefixes_path)

    default_namespace_prefixes = {
        prefix: ns for prefix, ns in zip(default_namespace_prefixes, default_namespaces)
    }
    prefixes.update(default_namespace_prefixes)

    if onto_ref_folder:
        file_prefixes = iterate_ttl_graphs(onto_ref_folder, read_prefixes_from_graph)
        prefixes |= merge_dictionaries(file_prefixes)
        inv_prefixes = {value: key for key, value in prefixes.items()}

        residual_file_prefixes = iterate_ttl_graphs(
            onto_ref_folder,
            partial(generate_residual_prefixes, inv_prefixes=inv_prefixes),
        )
        residual_file_prefixes = {
            key: value
            for residual_prefixes in residual_file_prefixes
            for key, value in residual_prefixes.items()
        }
        prefixes.update(residual_file_prefixes)
        inv_prefixes = {value: key for key, value in prefixes.items()}

    search_terms = get_search_terms_from_defaults(default_namespace_prefixes)

    if onto_ref_folder:
        file_search_terms = iterate_ttl_graphs(
            onto_ref_folder,
            partial(get_search_terms_from_graph, inv_prefixes=inv_prefixes),
        )
        search_terms |= merge_dictionaries(file_search_terms)

    aliases = {
        term: aliases
        for term, aliases in iter_diagram_terms(
            graph, lambda term: (term, get_term_aliases(term))
        )
    }

    literal_terms = {
        term
        for term in filter(lambda term: ('"' in term), get_diagram_terms_iter(graph))
    }
    constructed_terms = {
        term: term_uri_ref
        for term, term_uri_ref in map(
            lambda term_info: (
                term_info[0],
                construct_term_uri(
                    *get_abbrev_term(term_info[0], term_info[1]), prefixes=prefixes
                ),
            ),
            filter(
                lambda term_info: term_info[0] not in literal_terms,
                get_diagram_terms_iter_with_pred(graph),
            ),
        )
    }
    search_keys = {
        term: search_key
        for term, search_key in iter_diagram_terms(
            graph, lambda term: (term, get_term_search_keys(term, inv_prefixes))
        )
    }
    substitution_results = {
        term: substituted_value
        for term, substituted_value in iter_diagram_terms(
            graph,
            lambda term: (term, substitute_term(search_keys[term], search_terms)),
        )
        if substituted_value is not None
    }

    inv_constructed_terms = {value: key for key, value in constructed_terms.items()}

    constructed_terms.update(substitution_results)
    constructed_literal_terms = {
        term: construct_literal(
            term, lang=get_literal_lang_annotation(term), datatype=get_literal_data_type(term, search_terms)
        )
        for term in literal_terms
    }
    constructed_terms.update(constructed_literal_terms)

    output_graph = nx.DiGraph()
    for subj, obj, data in graph.edges(data=True):
        pred = data["label"]
        subj, obj, pred = tuple(constructed_terms[key] for key in (subj, obj, pred))
        output_graph.add_edge(subj, obj, label=pred)

    class_terms = get_class_terms(output_graph)
    predicate_terms = {data["label"] for _, _, data in output_graph.edges(data=True)}
    literal_terms = set(constructed_literal_terms.values())
    class_terms -= predicate_terms
    all_terms = (output_graph.nodes() | predicate_terms) - literal_terms

    pred_doms_ranges = get_doms_ranges(output_graph)

    # # create the rdf graph to store the ttl output
    rdf_graph = rdflib.Graph()

    # bind prefixes to namespaces for the rdf graph
    rdf_graph = bind_prefixes(rdf_graph, prefixes)

    # add all of the class terms as a type
    for term in class_terms:
        rdf_graph.add((term, RDF.type, OWL.Class))

    # if the term is a predicate and is not part of the default namespaces, add an object property type to the ttl file
    for term in predicate_terms:
        # TODO: Assume all predicates are object properties for now, change later
        if term_not_in_default_namespace(
            term, inv_prefixes, default_namespace_prefixes
        ):
            rdf_graph.add((term, RDF.type, OWL.ObjectProperty))

    term_in_search_results_filter = partial(
        term_in_search_results, inv_prefixes=inv_prefixes, search_terms=search_terms
    )
    term_not_in_default_namespace_filter = partial(
        term_not_in_default_namespace,
        inv_prefixes=inv_prefixes,
        default_namespace_prefixes=default_namespace_prefixes,
    )

    exact_match_property_predicates = [RDF.value, RDFS.label]
    exact_match_properties = {
        term: {prop: value}
        for prop in exact_match_property_predicates
        for result in map(
            lambda rdf_graph, prop=prop: map(
                lambda graph_term: (
                    graph_term,
                    get_term_value(subj=graph_term, pred=prop, ref_rdf_graph=rdf_graph),
                ),
                all_terms,
            ),
            get_ttl_file_iter(onto_ref_folder),
        )
        for term, value in result
    }
    rdf_graph = reduce(
        lambda rdf_graph, graph_term: add_exact_matches(
            term=graph_term,
            match_properties=exact_match_properties[graph_term],
            rdf_graph=rdf_graph,
        ),
        filter(
            term_in_search_results_filter,
            filter(term_not_in_default_namespace_filter, all_terms),
        ),
        rdf_graph,
    )
    rdf_graph = reduce(
        lambda rdf_graph, graph_term: add_labels(
            term=graph_term,
            labels=aliases[inv_constructed_terms[graph_term]],
            rdf_graph=rdf_graph,
        ),
        filter(
            term_not_in_default_namespace_filter,
            filterfalse(term_in_search_results_filter, all_terms),
        ),
        rdf_graph,
    )
    rdf_graph = reduce(
        lambda rdf_graph, graph_term: add_domains_ranges(
            term=graph_term,
            domains_ranges=pred_doms_ranges[graph_term],
            rdf_graph=rdf_graph,
        ),
        filter(
            term_not_in_default_namespace_filter,
            filterfalse(term_in_search_results_filter, predicate_terms),
        ),
        rdf_graph,
    )

    # now add the triples from the drawio diagram
    for domain_term, range_term, data in output_graph.edges(data=True):
        predicate_term = data["label"]
        rdf_graph.add((domain_term, predicate_term, range_term))
    # serialize the output as a turtle file
    rdf_graph.serialize(output_path)



from pathlib import Path

from networkx import DiGraph

from cemento.draw_io.constants import DiagramKey
from cemento.draw_io.transforms import (
    extract_elements,
    generate_graph,
    parse_elements,
    relabel_graph_nodes,
)


def read_drawio(
    input_path: str | Path,
    relabel_key: DiagramKey = DiagramKey.LABEL,
    inverted_rank_arrow: bool = False,
) -> DiGraph:
    elements = parse_elements(input_path)
    term_ids, rel_ids = extract_elements(elements)
    rank_terms = ["rdfs:subClassOf", "rdf:type"]
    graph = generate_graph(
        elements, term_ids, rel_ids, rank_terms, inverted_rank_arrow=inverted_rank_arrow
    )
    graph = relabel_graph_nodes(graph, new_attr_label=relabel_key.value)
    return graph
